import subprocess
import os
import psutil
import threading as t
import fysom as f
import instana.log as l
import instana.agent_const as a


class Discovery(object):
    pid = 0
    name = None
    args = None

    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class Fsm(object):
    RETRY_PERIOD = 30

    agent = None
    fsm = None
    timer = None

    def __init__(self, agent):
        l.info("Stan is on the scene.  Starting Instana instrumentation.")
        l.debug("initializing fsm")

        self.agent = agent
        self.fsm = f.Fysom({
            "initial": "lostandalone",
            "events": [
                ("startup",  "*",            "lostandalone"),
                ("lookup",   "lostandalone", "found"),
                ("announce", "found",        "announced"),
                ("ready",    "announced",    "good2go")],
            "callbacks": {
                "onlookup":       self.lookup_agent_host,
                "onannounce":     self.announce_sensor,
                "onchangestate":  self.printstatechange}})

    def printstatechange(self, e):
        l.debug('========= (%i#%s) FSM event: %s, src: %s, dst: %s ==========' % \
                (os.getpid(), t.current_thread().name, e.event, e.src, e.dst))

    def reset(self):
        self.fsm.lookup()

    def lookup_agent_host(self, e):
        if self.agent.sensor.options.agent_host != "":
            host = self.agent.sensor.options.agent_host
        else:
            host = a.AGENT_DEFAULT_HOST

        h = self.check_host(host)
        if h == a.AGENT_HEADER:
            self.agent.set_host(host)
            self.fsm.announce()
        else:
            host = self.get_default_gateway()
            if host:
                self.check_host(host)
                if h == a.AGENT_HEADER:
                    self.agent.set_host(host)
                    self.fsm.announce()
            else:
                l.error("Cannot lookup agent host. Scheduling retry.")
                self.schedule_retry(self.lookup_agent_host, e, "agent_lookup")
                return False
        return True

    def get_default_gateway(self):
        l.debug("checking default gateway")

        try:
            proc = subprocess.Popen(
                "/sbin/ip route | awk '/default/ { print $3 }'", stdout=subprocess.PIPE)

            return proc.stdout.read()
        except Exception as e:
            l.error(e)

            return None

    def check_host(self, host):
        l.debug("checking host", host)

        (_, h) = self.agent.request_header(
            self.agent.make_host_url(host, "/"), "GET", "Server")

        return h

    def announce_sensor(self, e):
        l.debug("announcing sensor to the agent")
        p = psutil.Process(os.getpid())
        d = Discovery(pid=p.pid,
                      name=p.cmdline()[0],
                      args=p.cmdline()[1:])

        (b, _) = self.agent.request_response(
            self.agent.make_url(a.AGENT_DISCOVERY_URL), "PUT", d)
        if not b:
            l.error("Cannot announce sensor. Scheduling retry.")
            self.schedule_retry(self.announce_sensor, e, "announce")
            return False
        else:
            self.agent.set_from(b)
            self.fsm.ready()
            l.warn("Host agent available. We're in business. (Announced pid: %i)" % p.pid)
            return True

    def schedule_retry(self, fun, e, name):
        l.debug("Scheduling: " + name)
        self.timer = t.Timer(self.RETRY_PERIOD, fun, [e])
        self.timer.daemon = True
        self.timer.name = name
        self.timer.start()
        l.debug('Threadlist: %s', str(t.enumerate()))

    def test_agent(self, e):
        l.debug("testing communication with the agent")

        (b, _) = self.agent.head(self.agent.make_url(a.AGENT_DATA_URL))

        if not b:
            self.schedule_retry(self.test_agent, e, "agent test")
        else:
            self.fsm.test()



import opentracing as ot
from instana import tracer
import opentracing.ext.tags as ext


class InstanaMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response
        ot.global_tracer = tracer.InstanaTracer()
        self

    def __call__(self, request):
        env = request.environ
        if 'HTTP_X_INSTANA_T' in env and 'HTTP_X_INSTANA_S' in env:
            ctx = ot.global_tracer.extract(ot.Format.HTTP_HEADERS, env)
            span = ot.global_tracer.start_span("django", child_of=ctx)
        else:
            span = ot.global_tracer.start_span("django")

        span.set_tag(ext.HTTP_URL, env['PATH_INFO'])
        span.set_tag("http.params", env['QUERY_STRING'])
        span.set_tag(ext.HTTP_METHOD, request.method)
        span.set_tag("http.host", env['HTTP_HOST'])

        response = self.get_response(request)

        span.set_tag(ext.HTTP_STATUS_CODE, response.status_code)
        ot.global_tracer.inject(span.context, ot.Format.HTTP_HEADERS, response)
        span.finish()
        return response



from __future__ import absolute_import
import opentracing as ot
from instana import util

prefix_tracer_state = 'HTTP_X_INSTANA_'
field_name_trace_id = prefix_tracer_state + 'T'
field_name_span_id = prefix_tracer_state + 'S'
field_count = 2


class HTTPPropagator():
    """A Propagator for Format.HTTP_HEADERS. """

    def inject(self, span_context, carrier):
        carrier[field_name_trace_id] = util.id_to_header(span_context.trace_id)
        carrier[field_name_span_id] = util.id_to_header(span_context.span_id)

    def extract(self, carrier):  # noqa
        count = 0
        span_id, trace_id = (0, 0)
        for k in carrier:
            v = carrier[k]
            k = k.lower()
            if k == field_name_span_id:
                span_id = util.header_to_id(v)
                count += 1
            elif k == field_name_trace_id:
                trace_id = util.header_to_id(v)
                count += 1

        if count != field_count:
            raise ot.SpanContextCorruptedException()

        return ot.SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            baggage={},
            sampled=True)



"""
Instana sensor and tracer. It consists of two modules that can be used as entry points:

- sensor: activates the meter to collect and transmit all kind of built-in metrics
- tracer: OpenTracing tracer implementation. It implicitly activates the meter
"""

__author__ = 'Instana Inc.'
__copyright__ = 'Copyright 2016 Instana Inc.'
__credits__ = ['Pavlo Baron']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Pavlo Baron'
__email__ = 'pavlo.baron@instana.com'

__all__ = ['sensor', 'tracer']
